/************************************************************************/
/*  Program reads a byte data cube for marching cubes
/*  and send those to the Scivis server
/************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h> 
#include <errno.h> 
#include <string.h> 
#include <sys/types.h> 
#include <sys/stat.h> 
#include <netinet/in.h> 
#include <sys/socket.h> 
#include <sys/wait.h> 

#include "march.h"
#include "java_ser.h"
#include "binio.h"

#define MYPORT 7655    /* the port users will be connecting to */
#define BACKLOG 10     /* how many pending connections queue will hold */
#define FILEWRITE 1    /* write received data to a file */
#define FACTOR   1000000000

#define PI         3.1415926535
#define TWOPI      (PI * 2.0)
#define PI8        (PI / 8.0)
#define PI16       (PI / 16.0)
#define PI32       (PI / 32.0)
#define PI64       (PI / 64.0)
#define PI128      (PI / 128.0)
#define PI256      (PI / 256.0)
#define PI512      (PI / 512.0)

#define dot_prod(p,q) ((p)[0]*(q)[0]+(p)[1]*(q)[1]+(p)[2]*(q)[2])

#define MAX_DATA_SIZE  50
#define SIZE_PER_DATA  200
#define DATA_LIST struct data_list

/*******************************************
 * contains whole data in cache
 *******************************************/
DATA_LIST {
   char title[256];
   int used;
   int ntimes;
   float time_value[SIZE_PER_DATA];
} node[MAX_DATA_SIZE];


/*******************************************
 * contain one time data in cache
 *******************************************/
float ***grid;
float *gridinfo;
unsigned  xsize, ysize, zsize;

int ntime;
float threshold;
float timeval;

/*******************************************
 * misc
 *******************************************/
int debug = 1;
int sockfd;     /* socket id to Scivis server*/
int data_version = 0;

int name_length;
char filename[256];


/********************************************
 * converting a positive integer into ASCII *
 ********************************************/
char* itoa(n) 
int n;
{
  int i, idx = 0;
  char  buf[100], *p;

  do {
    buf[idx++] = n % 10 + 48;
    n /= 10;
  } while (n);
  p = (char *)malloc(idx + 1);
  for (i=0; i<idx; i++)
    p[i] = buf[idx - i - 1];
  p[idx] = '\0';
  return p;
}



/********************************************
 * search nodes and return position
 ********************************************/
DATA_LIST *search(name) 
char *name;
{
   DATA_LIST *p;
   int index = 0;

   for (p=node; p<node+MAX_DATA_SIZE; p++) {
     if (p->used == 1) { 
       if (strcmp(p->title, name) == 0) {
          data_version = index;
          return p;
       }
     }
     index++; 
   }
   return 0;
}

/************************************************************************
 * send triangle information 
 ************************************************************************/
int send_tri(num_connections, num_corners, corner_coords, connections, temp_count,corner_norms, threshold, zmin, zmax)
 Big connections[];
 Output *corner_coords[3],*corner_norms[3];
 Big temp_count;
 long num_connections, num_corners;
 float threshold, zmin, zmax;
{

  int n = 0;
  long j = 3;
  int tmp1, tmp2, tmp3;
  float temp1, temp2, temp3;
  int  size;

   write_float4(sockfd, timeval);
   write_int4(sockfd, xsize);
   write_int4(sockfd, ysize);
   write_int4(sockfd, zsize);

   write_float4(sockfd, threshold);
   write_float4(sockfd, zmin);
   write_float4(sockfd, zmax);

/********************************************************/ 
/* write coordinate points to output_fp */

  size = num_corners;
  write_int4(sockfd, size);
  
  for(n=1; n<=num_corners; n++) {
     temp1 = corner_coords[0][n];
     temp2 = corner_coords[1][n];
     temp3 = corner_coords[2][n];
     write_float4(sockfd, temp1);
     write_float4(sockfd, temp2);
     write_float4(sockfd, temp3);
  } 


/*******************************************************/      
/* write face connectivity */
  size = num_connections;
  write_int4(sockfd, size);

  for (n = 0; n<num_connections; n++ ) {
       tmp1 = connections[n]-1;
       tmp2 = connections[n+1]-1;
       tmp3 = connections[n+2]-1;
       write_int4(sockfd, tmp1);
       write_int4(sockfd, tmp2);
       write_int4(sockfd, tmp3);
       n+=2;
   }     

  return 0;

} /* send */


/*************************************************************
 *************************************************************/
calc_iso() {
   int i,j,k,l,iz, index, pos;
   float zmin = 1.0e10f, zmax = 1.0e-10f;

   /*  marching cubes variables  */
   Big p_edges, temp_count, march();
   long num_connections, num_corners;
   Big *connections;
   double *corner_coords[3], *corner_norms[3];
   int return_val;

   /*********************************************************
    *  make grid and calc isosurface
    *********************************************************/

   if((grid = (float ***) Calloc(xsize, sizeof(float **))) == NULL) 
     perror("grid"); 
   if((grid[0] = (float **) Calloc(xsize*ysize, sizeof(float *))) == NULL) 
     perror("grid[]"); 
   if((grid[0][0] = (float *) Calloc(xsize*ysize*zsize, sizeof(float))) == NULL) 
     perror("grid[][]"); 

   for(i = 0; i < xsize; i++) {
       grid[i] = *grid + i*ysize;
       for(j = 0; j < ysize; j++) 
	 grid[i][j] = **grid + ((ysize*i)+j)*zsize;  
   }

   /* initialize first slice to zero inorder to get rid of edges extending   */
   /* to extents of the working space                                        */
   for(i=0;i<xsize;i++)
     for(j=0;j<ysize;j++)
       grid[i][j][0] = 0.0;


   index = 0;
   for(iz=0; iz<zsize; iz++) 
     for(i=0; i<ysize; i++) 
        for(j=0; j<xsize; j++) {
           grid[j][i][iz] = gridinfo[index];
           index++;
        }

   for(iz=0; iz<zsize; iz++) 
     for(i=0; i<ysize; i++) 
        for(j=0; j<xsize; j++) {
           if (zmin > grid[j][i][iz]) zmin = grid[j][i][iz];
           if (zmax < grid[j][i][iz]) zmax = grid[j][i][iz];
       }


   /***************************************************************
    * marching cube
    ***************************************************************/

if (debug) {   
printf("\ncalc for time value %f\n", timeval);
}

   temp_count = (Big)march( (unsigned long)xsize,
                 (unsigned long)ysize, (unsigned long)zsize,
                 (Input)threshold, (Input ***)grid, 
                 corner_coords, corner_norms,
                 &connections, &num_corners, &num_connections, &p_edges);

 
   sockfd = initialize();
   if (sockfd == -1) {
     printf("Can't connect the Scivis server.\n");
     return 1;
   }

   /* first send command(13) */
   write_int4(sockfd, 13);

   /* send port */
   write_int4(sockfd, MYPORT);

   /* send title information */
   name_length = strlen(filename);
   write_int4(sockfd, name_length);
   write(sockfd, &filename, name_length);

   /* send data version */
   write_int4(sockfd, data_version);

   /* copy the second row to the first row */
   send_tri(num_connections,num_corners,corner_coords, connections, 
           temp_count,corner_norms, (Input)threshold, zmin, zmax);
}



/*************************************************************
 * action when this server receives data set from the client
 *************************************************************/
do_type_one(new_fd)
int new_fd;
{
   int i, j, size, dummy, end_process, xs, ys, zs;
   DATA_LIST *search(), *pnode = node;
   char tmpname[256], tmpname2[256];

   int handle, whandle;

   /* first receive title information */
   read_int4(new_fd, &name_length);
   receive_data(new_fd, filename, name_length);
   filename[name_length] = '\0';

   /* receive the number of time and initial threshold */
   read_int4(new_fd, &ntime);
   read_float4(new_fd, &threshold); 

    /* receive remaining information */
   for (i=0; i<ntime; i++) {
      read_float4(new_fd, &timeval); 
      read_int4(new_fd, &xs);
      read_int4(new_fd, &ys);
      read_int4(new_fd, &zs);
      xsize = xs;
      ysize = ys;
      zsize = zs;

      size = xsize * ysize * zsize;

      if (gridinfo != 0)  free(gridinfo);
      gridinfo = calloc(size, sizeof(float)); 

      for (j=0; j<size; j++) 
         read_float4(new_fd, &gridinfo[j]);


      /**************************************************
       * handle after reading whole data
       **************************************************/
      end_process = 0; 

      if (pnode = search(filename)) {      /* Same title data is exists */

        for(j=0; j<pnode->ntimes; j++) {

          /* CASE 1: Same title data and same time */
	  if (pnode->time_value[j] == timeval) { 
             pnode->time_value[j] = timeval;
             end_process = 1;

             dummy = (data_version+1) * 1000 + j;

             break;
          }
        }

        /* CASE 2: Same title data but different time */
        if (end_process != 1) {
           dummy = (data_version+1) * 1000 + pnode->ntimes;
           pnode->time_value[pnode->ntimes] = timeval;
           pnode->ntimes++;
        } 


      /* CASE 3: new time data */
      } else {
        for (j=0; j<MAX_DATA_SIZE; j++) {
          if (node[j].used != 1) {
             data_version = j;
             break;
          }
        }
                            
        size = strlen(filename);
        for (j=0; j<size; j++)
           node[data_version].title[j] = filename[j];
        node[data_version].title[j] = '\0';

        node[data_version].ntimes = 1;
        node[data_version].time_value[0] = timeval;
        node[data_version].used = 1;

        dummy = (data_version+1) * 1000;
      }

      strcpy(tmpname, "cache/");
      strncat(tmpname, itoa(dummy), strlen(itoa(dummy)));

      if ((handle = creat(tmpname, S_IREAD | S_IWRITE)) < 0) {
         perror("Can't create cache file.");
         return;
      }

      strcpy(tmpname2, "cache/");
      strncat(tmpname2, itoa(dummy), strlen(itoa(dummy)));
      strncat(tmpname2, ".dat", strlen(".dat"));

      if ((whandle = creat(tmpname2, S_IREAD | S_IWRITE)) < 0) {
         perror("Can't create data file.");
         return;
      }

      /**************************************************
       * Write the data to a file
       **************************************************/
      write_float4(handle, timeval);    /* double check */
      write_float4(whandle, timeval);

      /* write title information */
      name_length = strlen(filename);
      write_int4(handle, name_length);
      write(handle, &filename, name_length);

      write_int4(handle, xsize);
      write_int4(handle, ysize);
      write_int4(handle, zsize);
      write_int4(whandle, xsize);
      write_int4(whandle, ysize);
      write_int4(whandle, zsize);

      size = xsize * ysize * zsize;
      for (j=0; j<size; j++) {
         write_float4(handle, gridinfo[j]);
         write_float4(whandle, gridinfo[j]);
      }

      close(handle);
      close(whandle);
      calc_iso();
   }
}

/*************************************************************
 * action when this server receives "exit server" command
 *************************************************************/
do_type_two()
{
  /* clear cache */
  system("rm -r cache");
  printf("\n\n *** Isoserver exits ***\n\n");
}



/*************************************************************
 * action when this server receives "change threshold" command
 *************************************************************/
do_type_three(new_fd)
int new_fd;
{

    int i, j, onlyone, dummy, xs, ys, zs;
    int handle;    /* file handle */
    float this_timeval;
    
    char tmpname[256];

    read_int4(new_fd, &data_version);
    read_int4(new_fd, &onlyone);
 
    if (onlyone == 1) 
       read_float4(new_fd, &timeval);

    read_float4(new_fd, &threshold);

    for (i=0; i<node[data_version].ntimes; i++) {
      if (onlyone == 1) {
	if (node[data_version].time_value[i] != timeval) 
            continue;
      }

      dummy = (data_version+1) * 1000 + i;
      strcpy(tmpname, "cache/");
      strncat(tmpname, itoa(dummy), strlen(itoa(dummy)));

      if ((handle = open(tmpname, O_RDONLY)) < 0) {
         perror("Can't open cache file.");
         return;
      }

      /**************************************************
       * read the data from a file
       **************************************************/
      read_float4(handle, &this_timeval);    /* double check */
      if (node[data_version].time_value[i] != this_timeval) {
         perror("Incorrect cache data...");
         return;
      }
      timeval = this_timeval;

      /* and then receive title information */
      read_int4(handle, &name_length);
      receive_data(handle, filename, name_length);
      filename[name_length] = '\0';

      read_int4(handle, &xs);
      read_int4(handle, &ys);
      read_int4(handle, &zs);
      xsize = xs;
      ysize = ys;
      zsize = zs;

      dummy = xsize * ysize * zsize;
      for (j=0; j<dummy; j++) 
         read_float4(handle, &gridinfo[j]);

      close(handle);
      calc_iso();
    }
}


/*************************************************************
 * action when this server receives "clear cache" command
 *************************************************************/
do_type_four(new_fd)
int new_fd;
{

    int i, dummy;
    char tmpname[256], tmpname2[256];

    read_int4(new_fd, &data_version);

    for (i=0; i<node[data_version].ntimes; i++) {
      dummy = (data_version+1) * 1000 + i;
      strcpy(tmpname, "cache/");
      strncat(tmpname, itoa(dummy), strlen(itoa(dummy)));
      unlink(tmpname);
printf("remove %s\n", tmpname);

      strcpy(tmpname2, "cache/");
      strncat(tmpname2, itoa(dummy), strlen(itoa(dummy)));
      strncat(tmpname2, ".dat", strlen(".dat"));
      unlink(tmpname2);
printf("remove %s\n", tmpname2);
    }

    node[data_version].used = 0;
}


/*************************************************************
 * extract the grid information and send it to Scivis
 *************************************************************/
do_type_five(new_fd)
int new_fd;
{
   int i, j, k, orient, extraction, f_data_version, f_name_length;
   int xs, ys, zs, dummy;
   int handle;    /* file handle */
   float this_timeval, fdummy;
   char tmpname[256], f_name[256];
 
   read_int4(new_fd, &f_data_version);

   /* receive data set name */
   read_int4(new_fd, &f_name_length);
   receive_data(new_fd, f_name, f_name_length);
   f_name[f_name_length] = '\0';

   read_int4(new_fd, &orient);
   read_int4(new_fd, &extraction);

   for (i=0; i<node[f_data_version].ntimes; i++) {

      sockfd = initialize();

      /* first send command(34) */
      write_int4(sockfd, 34);
   
      /* send data set name */
      f_name_length = strlen(f_name);
      write_int4(sockfd, f_name_length);
      write(sockfd, &f_name, f_name_length);

      dummy = (f_data_version+1) * 1000 + i;
      strcpy(tmpname, "cache/");
      strncat(tmpname, itoa(dummy), strlen(itoa(dummy)));
      strncat(tmpname, ".dat", strlen(".dat"));

      if ((handle = open(tmpname, O_RDONLY)) < 0) {
         perror("Can't open data file.");
         return;
      }

      /**************************************************
       * read the data from a file
       **************************************************/
      read_float4(handle, &this_timeval);
      write_float4(sockfd, this_timeval);

      read_int4(handle, &xs);
      read_int4(handle, &ys);
      read_int4(handle, &zs);
      write_int4(sockfd, orient);
      write_int4(sockfd, extraction);

      if (orient == 0) {                /* X Axis extraction */

          dummy = ys * zs;
          lseek(handle, extraction*dummy*4, SEEK_CUR);  /* SEEK_CUR = 1 */

          write_int4(sockfd, dummy);
          for (j=0; j<dummy; j++) {
             read_float4(handle, &fdummy);
             write_float4(sockfd, fdummy);
          }

      } else if (orient == 1) {         /* Y Axis extraction */

          dummy = xs * zs;

          write_int4(sockfd, dummy);
          for (j=0; j<xs; j++) {
             lseek(handle, (zs*extraction)*4, SEEK_CUR); 

             for (k=0; k<zs; k++) {
                read_float4(handle, &fdummy);
                write_float4(sockfd, fdummy);
             }
             lseek(handle, zs*(ys-extraction-1)*4, SEEK_CUR); 
          }

      } else if (orient == 2) {         /* Z Axis extraction */
          dummy = xs * ys;

          write_int4(sockfd, dummy);
          for (j=0; j<xs; j++) {
             for (k=0; k<ys; k++) {
                lseek(handle, extraction*4, SEEK_CUR); 
                read_float4(handle, &fdummy);
                write_float4(sockfd, fdummy);
                lseek(handle, (zs-extraction-1)*4, SEEK_CUR); 
             }
          }
      }

      close(handle);
    }
    close(sockfd);
}

/*************************************************************
 * send all data to Scivis
 *************************************************************/
do_type_six(new_fd)
int new_fd;
{

   int i, j, xs, ys, zs, dummy, f_name_length, f_data_version;
   int handle;    /* file handle */
   float this_timeval, fdummy;
   char tmpname[256], f_name[256];
 
   sockfd = initialize();

   read_int4(new_fd, &f_data_version);

   /* receive file name */
   read_int4(new_fd, &f_name_length);
   receive_data(new_fd, f_name, f_name_length);
   f_name[f_name_length] = '\0';

   /* first send command(24) */
   write_int4(sockfd, 24);
   
   /* send file name */
   f_name_length = strlen(f_name);
   write_int4(sockfd, f_name_length);
   write(sockfd, &f_name, f_name_length);

   write_int4(sockfd, node[f_data_version].ntimes);

   for (i=0; i<node[f_data_version].ntimes; i++) {

      dummy = (f_data_version+1) * 1000 + i;
      strcpy(tmpname, "cache/");
      strncat(tmpname, itoa(dummy), strlen(itoa(dummy)));
      strncat(tmpname, ".dat", strlen(".dat"));

      if ((handle = open(tmpname, O_RDONLY)) < 0) {
         perror("Can't open data file.");
         return;
      }

      /**************************************************
       * read the data from a file
       **************************************************/
      read_float4(handle, &this_timeval);
      write_float4(sockfd, this_timeval);

      read_int4(handle, &xs);
      read_int4(handle, &ys);
      read_int4(handle, &zs);
      write_int4(sockfd, xs);
      write_int4(sockfd, ys);
      write_int4(sockfd, zs);

      dummy = xs * ys * zs;
      for (j=0; j<dummy; j++) {
         read_float4(handle, &fdummy);
         write_float4(sockfd, fdummy);
      }

      close(handle);
    }
    close(sockfd);
}



/*************************************************************
 * This server doesn't use fork() function ....
 * Have to fix for parallel work
 *************************************************************/
main() {
    int sock_fd, new_fd;  /* listen on sock_fd, new connection on new_fd */
    struct sockaddr_in my_addr;    /* my address information */
    struct sockaddr_in their_addr; /* connector's address information */
    int sin_size;
    struct stat statbuf;

    int todo, i;

    /* before start the server, first make cache directory */
    stat("cache", &statbuf);

    if (statbuf.st_mode & S_IFDIR) 
       printf("Cache directory exists ...\n");
    else
       system("mkdir cache");

    /* clear used flag */
    for (i=0; i<MAX_DATA_SIZE; i++) 
      node[i].used =0;
 
    if ((sock_fd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
        perror("socket");
        return 1;
    }

    my_addr.sin_family = AF_INET;         /* host byte order */
    my_addr.sin_port = htons(MYPORT);     /* short, network byte order */
    my_addr.sin_addr.s_addr = INADDR_ANY; /* auto-fill with my IP */
    bzero(&(my_addr.sin_zero), 8);        /* zero the rest of the struct */

    if (bind(sock_fd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) \
                                                                  == -1) {
        perror("bind");
        return 1;
    }

    if (listen(sock_fd, BACKLOG) == -1) {
        perror("listen");
        return 1;
    }

    while(1) {  /* main accept() loop */
        sin_size = sizeof(struct sockaddr_in);
        if ((new_fd = accept(sock_fd, (struct sockaddr *)&their_addr, \
                                                      &sin_size)) == -1) {
            perror("accept");
            continue;
        }
        printf("\nserver: got connection from %s\n", \
                                           inet_ntoa(their_addr.sin_addr));
            
        read_int4(new_fd, &todo);

        /**********************************************************/
        if (todo == 1) {            /* received data */
           do_type_one(new_fd);

        } else if (todo == 2) {     /* exit server */
           close(new_fd); 
           do_type_two();
           return 1;

        } else if (todo == 3) {     /* recalculate isosurface */
           do_type_three(new_fd);

        } else if (todo == 4) {     /* clear cache command */
           do_type_four(new_fd);

        } else if (todo == 5) {     /* extract grid data */
           do_type_five(new_fd);

        } else if (todo == 6) {     /* send all data to Scivis */
           do_type_six(new_fd);
        }
        /**********************************************************/

        close(new_fd); 
    }
} /* end of main */

